 /**
  * stream_server.c:
  * OpenCV video streaming server
  *
  * Author  Nash <me_at_nashruddin.com>
  *         gartim gartim_at_gmail.com
  *
  * todo:
  * - clean up port assignments for select process
  * - change from chat broadcast commend to sending to usb controller
  * - add motor to h/w and test
  *
  * See the tutorial at
  * http://nashruddin.com/StrEAMinG_oPENcv_vIdEos_ovER_tHe_nEtWoRk
  */
    
#define VIDEO_PORT 8888
#define CNTL_PORT  "8889"
#define CHANNEL_0  0
#define CHANNEL_1  1
#define CHANNEL_4  4
#define CHANNEL_5  5
    
#include <fcntl.h>
#include <termios.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include "cv.h"
#include "highgui.h"
// command-control
#include <sys/time.h>
#include <string.h>
#include <netdb.h>

  
  
CvCapture*  capture;
IplImage*   img0;
IplImage*   img1;
int         is_data_ready = 0;
int         serversock, clientsock;
struct      timeval tv;
  
pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
  
void* streamServer(void* arg);
void  quit(char* msg, int retval);
  

// Gets the position of a Maestro channel.
// See the "Serial Servo Commands" section of the user's guide.
int maestro_get_position(int fd, unsigned char channel)
{
	unsigned char command[] = {0x90, channel};
	if(write(fd, command, sizeof(command)) == -1) {
		perror("error writing");
		return -1;
	}
	unsigned char response[2];
	if(read(fd,response,2) != 2) {
		perror("error reading");
		return -1;
	}
   
	return response[0] + 256*response[1];

}
// Sets the target of a Maestro channel.
// See the "Serial Servo Commands" section of the user's guide.
// The units of 'target' are quarter-microseconds.
int maestro_set_target(int fd, unsigned char channel, unsigned short target)
{
	unsigned char command[] = {0x84, channel, target & 0x7F, target >> 7 & 0x7F};
	if (write(fd, command, sizeof(command)) == -1) {
		perror("error writing");
		return -1;
	}
	return 0;
}
 
int left_right_position = 1492;
int up_down_position = 1492;

int maestro_turn(int fd, char direction)
{
	// 992..2000(*4) or a range of 3968..8000
	int min = 992;
	int max = 2000;
	printf("Current direction(%c) position l/r, u/d (%d, %d)\n", direction, left_right_position, up_down_position) ;
	if (direction == 'l'  && left_right_position >= min) {
		left_right_position = left_right_position-25;
		maestro_set_target(fd, CHANNEL_0, (left_right_position*4));
	}
	else if (direction == 'r' &&  left_right_position <= max) {
		left_right_position = left_right_position+25;
		maestro_set_target(fd, CHANNEL_0, (left_right_position*4));
	}
	else if (direction == 'u' &&  up_down_position >= min) {
		up_down_position = up_down_position-25;
		maestro_set_target(fd, CHANNEL_1, (up_down_position*4));
	}
	else if (direction == 'd' &&  up_down_position <= max) {
		up_down_position = up_down_position+25;
		maestro_set_target(fd, CHANNEL_1, (up_down_position*4));
	}
	printf("Current direction(%c) position l/r, u/d (%d, %d)\n", direction, left_right_position, up_down_position);
}
int maestro_direction(int fd, char direction)
{
	// f = forward n degrees, b = back n degrees, F = forward-continuous, B = back-continuous
	// 0..1200(*4) backword,  1201..2400(*4) forward
	int i;
	if (direction == 'f') {
		printf("wheel direction forward\n") ;
		maestro_set_target(fd, CHANNEL_5, (1800*4));
		maestro_set_target(fd, CHANNEL_4, (992*4));
	}
	else if (direction == 'b') {
		printf("wheel direction backwords\n") ;
		maestro_set_target(fd, CHANNEL_5, (992*4));
		maestro_set_target(fd, CHANNEL_4, (1800*4));
	}
	else if (direction == 's') {
		printf("Current wheel direction(%i)\n", 0) ;
		maestro_set_target(fd, CHANNEL_5, (0));
		maestro_set_target(fd, CHANNEL_4, (0));
	}
	else if (direction == 'e') {  // forward left
		printf("wheel direction forward left\n") ;
		maestro_set_target(fd, CHANNEL_5, (1600*4));
		maestro_set_target(fd, CHANNEL_4, (992*4));
	}
	else if (direction == 'i') {  // forward right
		printf("wheel direction forward right\n") ;
		maestro_set_target(fd, CHANNEL_5, (1800*4));
		maestro_set_target(fd, CHANNEL_4, (1392*4));
	}
	else if (direction == 't') {
	        printf("slowly move in steps of 25...\n");
		for (i=992; i <= 3000; i=i+25) {
			printf("setting(%i) times4(%i)\n", i, i*4);
			maestro_set_target(fd, CHANNEL_4, (i*4));
			sleep(1);
		}
	}
}

void *get_in_addr(struct sockaddr *sa) {
	if (sa->sa_family == AF_INET) {
		return &(((struct sockaddr_in*)sa)->sin_addr);
	}
	return &(((struct sockaddr_in6*)sa)->sin6_addr);
}
int main(int argc, char** argv)
 {
	pthread_t   thread_s;
	int         key;
	int         fdnet, value, sock;
  
	if (argc == 2) {
		//capture = cvCaptureFromFile(argv[1]);
		int cam = atoi (argv[1]);
		capture = cvCaptureFromCAM(cam);
	} else {
		capture = cvCaptureFromCAM(1);
	}
  
	if (!capture) {
		quit("cvCapture failed", 1);
	}

	cvSetCaptureProperty(capture, CV_CAP_PROP_FRAME_WIDTH, 320);
	cvSetCaptureProperty(capture, CV_CAP_PROP_FRAME_HEIGHT, 280);
  
	// Open the Maestro's virtual COM port.
	const char * device = "/dev/ttyACM0";  // Linux
	int indx, target, position;
	int fd = open(device, O_RDWR | O_NOCTTY);
	if (fd == -1) {
		perror(device);
		return 1;
	}
	maestro_set_target(fd, 0, (1492*4));
	maestro_set_target(fd, 1, (1492*4));
	maestro_set_target(fd, 4, (1492*4));
	maestro_set_target(fd, 5, (1492*4));
	// setup select process for commands

	fd_set master;    // master file descriptor list
	fd_set read_fds;  // temp file descriptor list for select()
	int fdmax;        // maximum file descriptor number
	int listener;     // listening socket descriptor
	int newfd;        // newly accept()ed socket descriptor
	struct sockaddr_storage remoteaddr; // client address
	socklen_t addrlen;
	char buf[256];    // buffer for client data
	char remoteIP[INET6_ADDRSTRLEN];
	int yes=1;        // for setsockopt() SO_REUSEADDR, below
	int i, j, rv, nbytes;
	struct addrinfo hints, *ai, *p;
	FD_ZERO(&master);    // clear the master and temp sets
	FD_ZERO(&read_fds);
	// get us a socket and bind it
	memset(&hints, 0, sizeof hints);
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	if ((rv = getaddrinfo(NULL, CNTL_PORT, &hints, &ai)) != 0) {
		fprintf(stderr, "stream_server: %s\n", gai_strerror(rv));
		exit(1);
	}
	for(p = ai; p != NULL; p = p->ai_next) {
		listener = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
		if (listener < 0) { 
			continue;
		}
		// lose the pesky "address already in use" error message
		setsockopt(listener, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int));
		if (bind(listener, p->ai_addr, p->ai_addrlen) < 0) {
			close(listener);
			continue;
		}
		break;
	}
	// if we got here, it means we didn't get bound
	if (p == NULL) {
		fprintf(stderr, "stream_server: failed to bind\n");
		exit(2);
	}
	freeaddrinfo(ai); // all done with this
	// listen
	if (listen(listener, 10) == -1) {
		perror("listen");
		exit(3);
	}
	// add the listener to the master set
	FD_SET(listener, &master);
	// keep track of the biggest file descriptor
	fdmax = listener; // so far, it's this one
  
	img0 = cvQueryFrame(capture);
	img1 = cvCreateImage(cvGetSize(img0), IPL_DEPTH_8U, 1);
  
	cvZero(img1);
	cvNamedWindow("stream_server", CV_WINDOW_AUTOSIZE);
  
	/* print the width and height of the frame, needed by the client */
	fprintf(stdout, "stream_server: width:  %d\nheight: %d\n\n", img0->width, img0->height);
	fprintf(stdout, "Press 'q' to quit.\n\n");
  
	/* run the streaming server as a separate thread */
	if (pthread_create(&thread_s, NULL, streamServer, NULL)) {
		quit("pthread_create failed.", 1);
	}
  
	while( 1 ) {
		tv.tv_sec = 0;
		tv.tv_usec = 100000;
		read_fds = master; // copy it
		if (select(fdmax+1, &read_fds, NULL, NULL, &tv) == -1) {
			perror("select");
			exit(4);
		}
		// run through the existing connections looking for data to read
		for(i = 0; i <= fdmax; i++) {
			if (FD_ISSET(i, &read_fds)) { // we got one!!
				if (i == listener) {
					// handle new connections
					addrlen = sizeof remoteaddr;
					newfd = accept(listener,
					 (struct sockaddr *)&remoteaddr,
					 &addrlen);

					if (newfd == -1) {
       						perror("accept");
					} else {
						FD_SET(newfd, &master); // add to master set
						if (newfd > fdmax) {    // keep track of the max
							fdmax = newfd;
						}
						printf("stream_server: new connection from %s on "
                                                  "socket %d\n",
                                                  inet_ntop(remoteaddr.ss_family,
                                                  get_in_addr((struct sockaddr*)&remoteaddr),
                                                  remoteIP, INET6_ADDRSTRLEN),
                                                  newfd);

					}
				} else {
					// handle data from a client
					if ((nbytes = recv(i, buf, sizeof buf, 0)) <= 0) {
						// got error or connection closed by client
						if (nbytes == 0) {
							// connection closed
							printf("stream_server: socket %d hung up\n", i);
						} else {
							perror("recv");
						}
						close(i); // bye!
						FD_CLR(i, &master); // remove from master set
					} else {
						// we got some data from a client
						printf("stream_server: got command: %s\n", buf);
						switch ( (int)buf[0] ) {
                        				case 'q':
								break;
                        				case 'r':
                                				printf("turning right\n");
								//maestro_turn_right(fd);
								maestro_turn(fd,'r');
                                				break;
                        				case 'l':
                                				printf("turning left\n");
								//maestro_turn_left(fd);
								maestro_turn(fd,'l');
                                				break;
                        				case 'u':
                                				printf("moving up\n");
								maestro_turn(fd,'u');
                                				break;
                        				case 'd':
                                				printf("moving down\n");
								maestro_turn(fd,'d');
                                				break;
                        				case 'f':
                                				printf("moving forward\n");
								maestro_direction(fd,'f');
                                				break;
                        				case 'b':
                                				printf("moving back\n");
								maestro_direction(fd,'b');
                                				break;
                        				case 'e':
                                				printf("moving forward left\n");
								maestro_direction(fd,'e');
                                				break;
                        				case 'i':
                                				printf("moving forward right\n");
								maestro_direction(fd,'i');
                                				break;
                        				case 's':
                                				printf("moving stop all motors\n");
								maestro_direction(fd,'s');
                                				break;
                        				case 't':
                                				printf("test loop on one motor\n");
								maestro_direction(fd,'t');
                                				break;
                        				case '?':
                                				printf("unknown option: %c\n", key);
                                				break;
						}
						//for(j = 0; j <= fdmax; j++) {
						//	// send to everyone!
						//	if (FD_ISSET(j, &master)) {
						//		// except the listener and ourselves
						//		if (j != listener && j != i) {
						//			if (send(j, buf, nbytes, 0) == -1) {
						//				perror("send");
						//			}
                                		//		}
                         			//	}
                        			//}
                    			}
                		} // END handle data from client
            		} // END got new incoming connection
        	} // END looping through file descriptors
		/* get a frame from camera */
		img0 = cvQueryFrame(capture);
		if (!img0) break;
  
		img0->origin = 0;
		/*
		cvFlip(img0, img0, -1);
		*/
  
		/**
		* convert to grayscale
		* note that the grayscaled image is the image to be sent to the client
		* so we enclose it with pthread_mutex_lock to make it thread safe
		*/
		pthread_mutex_lock(&mutex);
		cvCvtColor(img0, img1, CV_BGR2GRAY);
		is_data_ready = 1;
		pthread_mutex_unlock(&mutex);
		
		/* also display the video here on server */
		cvShowImage("stream_server", img0);
		key = cvWaitKey(3);
	}
  
	/* user has pressed 'q', terminate the streaming server */
	if (pthread_cancel(thread_s)) {
		("pthread_cancel failed.", 1);
	}
  
	/* free memory */
	cvDestroyWindow("stream_server");
	quit(NULL, 0);
}
  


/**
  * This is the streaming server, run as a separate thread
  * This function waits for a client to connect, and send the grayscaled images
  */
void* streamServer(void* arg)
{
     struct sockaddr_in server;
     int yes = 1;
  
     /* make this thread cancellable using pthread_cancel() */
     pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
     pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  
     /* open socket */
     if ((serversock = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
         quit("socket() failed", 1);
     }
  
     /* setup server's IP and port */
     memset(&server, 0, sizeof(server));
     server.sin_family = AF_INET;
     server.sin_port = htons(VIDEO_PORT);
     server.sin_addr.s_addr = INADDR_ANY;
  
     setsockopt(serversock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int));

     /* bind the socket */
     if (bind(serversock, (const void*)&server, sizeof(server)) == -1) {
         quit("bind() failed", 1);
     }
  
     /* wait for connection */
     if (listen(serversock, 10) == -1) {
         quit("listen() failed.", 1);
     }
  
     /* accept a client */
     if ((clientsock = accept(serversock, NULL, NULL)) == -1) {
         quit("accept() failed", 1);
     }
  
     /* the size of the data to be sent */
     int imgsize = img1->imageSize;
     int bytes, i;
  
     /* start sending images */
     while(1)
     {
         /* send the grayscaled frame, thread safe */
         pthread_mutex_lock(&mutex);
         if (is_data_ready) {
             bytes = send(clientsock, img1->imageData, imgsize, 0);
             is_data_ready = 0;
         }
         pthread_mutex_unlock(&mutex);
  
         /* if something went wrong, restart the connection */
         if (bytes != imgsize) {
             fprintf(stderr, "Connection closed.\n");
             close(clientsock);
  
             if ((clientsock = accept(serversock, NULL, NULL)) == -1) {
                 quit("accept() failed", 1);
             }
         }
  
         /* have we terminated yet? */
         pthread_testcancel();
  
         /* no, take a rest for a while */
         usleep(1000);
     }
}
  
/**
 * this function provides a way to exit nicely from the system
 */
void quit(char* msg, int retval)
{
     if (retval == 0) {
         fprintf(stdout, (msg == NULL ? "" : msg));
         fprintf(stdout, "\n");
     } else {
         fprintf(stderr, (msg == NULL ? "" : msg));
         fprintf(stderr, "\n");
     }
  
     if (clientsock) close(clientsock);
     if (serversock) close(serversock);
     if (capture) cvReleaseCapture(&capture);
     if (img1) cvReleaseImage(&img1);
  
     pthread_mutex_destroy(&mutex);
  
     exit(retval);
}
